MSG_TYPES = [
    'UNUSED',
    'Interface Configuration',
    'Reserved',
    'Zone Name',
    'Zone Status',
    'Zones Snapshot',
    'Partition Status',
    'Partitions Snapshot',
    'System Status',
    'X-10 Message Received',
    'Log Event',
    'Keypad Message Received',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Program Data Reply',
    'Reserved',
    'User Information Reply',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Command Failed',
    'Positive Acknowledge',
    'Negative Acknowledge',
    'Message Rejected',
    # 0x20
    'Reserved',
    'Interface Configuration Request',
    'Reserved',
    'Zone Name Request',
    'Zone Status Request',
    'Zones Snapshot Request',
    'Partition Status Request',
    'Partitions Snapshot Request',
    'System Status Request',
    'Send X-10 Message',
    'Log Event Request',
    'Send Keypad Text Message',
    'Keypad Terminal Mode REquest',
    'Reserved',
    'Reserved',
    'Reserved',
    'Program Data Request',
    'Program Data Command',
    'User Information Request with PIN',
    'User Information Request without PIN',
    'Set User Code Command with PIN',
    'Set User Code Command without PIN',
    'Set User Authorization Command with PIN',
    'Set User Authorization Command without PIN',
    'Reserved',
    'Reserved',
    'Store Communication Event Command',
    'Set Clock / Calendar Command',
    'Primary Keypad Function with PIN',
    'Primary Keypad Function without PIN',
    'Secondary Keypad Function',
    'Zone Bypass Toggle',
]


class Zone(object):
    STATUS_FLAGS = [
        'Faulted', 'Trouble', 'Bypass', 'Inhibit', 'Low battery',
        'Loss of supervision', 'Reserved',]

    TYPE_FLAGS = [
        ['Fire', '24 hour', 'Key-switch', 'Follower',
         'Entry / exit delay 1', 'Entry / exit delay 2',
         'Interior', 'Local only'],
        ['Keypad sounder', 'Yelping Siren', 'Steady Siren',
         'Chime', 'Bypassable', 'Group bypassable', 'Force armable',
         'Entry guard'],
        ['Fast loop response', 'Double EOL tamper', 'Trouble',
         'Cross zone', 'Dialer delay', 'Swinger shutdown',
         'Restorable', 'Listen in'],
    ]

    def __init__(self, number):
        self.number = number
        self.name = 'Unknown'
        self.state = None
        self.condition_flags = []
        self.type_flags = []

    @property
    def bypassed(self):
        return ('Inhibit' in self.condition_flags or
                'Bypass' in self.condition_flags)


class Partition(object):
    CONDITION_FLAGS = [
        ['Bypass code required', 'Fire trouble', 'Fire',
         'Pulsing buzzer', 'TLM fault memory', 'reserved',
         'Armed', 'Instant'],
        ['Previous alarm', 'Siren on', 'Steady siren on',
         'Alarm memory', 'Tamper', 'Cancel command entered',
         'Code entered', 'Cancel pending'],
        ['Reserved', 'Silent exit enabled', 'Entryguard (stay mode)',
         'Chime mode on', 'Entry', 'Delay expiration warning',
         'Exit 1', 'Exit 2'],
        ['LED extinguish', 'Cross timing', 'Recent closing being timed',
         'Reserved', 'Exit error triggered', 'Auto home inhibited',
         'Sensor low battery', 'Sensor lost supervision'],
        ['Zone bypassed', 'Force arm triggered by auto arm',
         'Ready to arm', 'Ready to force arm', 'Valid PIN accepted',
         'Chime on (sounding)', 'Error beep (triple beep)',
         'Tone on (activation tone)'],
        ['Entry 1', 'Open period', 'Alarm sent using phone number 1',
         'Alarm sent using phone number 2',
         'Alarm sent using phone number 3',
         'Cancel report is in the stack',
         'Keyswitch armed', 'Delay trip in progress (common zone)'],
    ]

    def __init__(self, number):
        self.number = number
        self.condition_flags = []
        self.last_user = None

    @property
    def armed(self):
        return 'Armed' in self.condition_flags


class System(object):
    STATUS_FLAGS = [
        ['Line seizure', 'Off hook', 'Initial handshake received',
         'Download in progress', 'Dialer delay in progress',
         'Using backup phone', 'Listen in active', 'Two way lockout'],
        ['Ground fault', 'Phone fault', 'Fail to communicate',
         'Fuse fault', 'Box tamper', 'Siren tamper/trouble', 'Low battery',
         'AC fail'],
        ['Expander box tamper', 'Expander AC failure',
         'Expander low battery', 'Expander loss of supervision',
         'Expander auxiliary output over current',
         'Auxiliary communication channel failure',
         'Expander bell fault', 'Reserved'],
        ['6 digit PIN enabled', 'Programming token in use',
         'PIN required for local download', 'Global pulsing buzzer',
         'Global Siren on', 'Global steady siren',
         'Bus device has line seized',
         'Bus device has requested sniff mode'],
        ['Dynamic battery test', 'AC power on', 'Low battery memory',
         'Ground fault memory', 'Fire alarm verification being timed',
         'Smoke power reset', '50 Hz line power detected',
         'Timing a high voltage battery charge'],
        ['Communication since last autotest', 'Power up delay in progress',
         'Walk test mode', 'Loss of system time', 'Enroll requested',
         'Test fixture mode', 'Control shutdown mode',
         'Timing a cancel window'],
        ['reserved', 'reserved', 'reserved', 'reserved', 'reserved',
         'reserved', 'reserved', 'Call back in progress'],
        ['Phone line faulted', 'Voltage present interrupt active',
         'House phone off hook', 'Phone line monitor enabled',
         'Sniffing', 'Last read was off hook', 'Listen in requested',
         'Listen in trigger'],
        ['Valid partition 1', 'Valid partition 2', 'Valid partition 3',
         'Valid partition 4', 'Valid partition 5', 'Valid partition 6',
         'Valid partition 7', 'Valid partition 8'],
    ]

    def __init__(self):
        self.panel_id = 0
        self.status_flags = []


class LogEvent(object):
    ZONE_EVENT_CODES = {
        0: 'Alarm',
        1: 'Alarm restore',
        2: 'Bypass',
        3: 'Bypass restore',
        4: 'Tamper',
        5: 'Tamper restore',
        6: 'Trouble',
        7: 'Trouble restore',
        8: 'TX low battery',
        9: 'TX low battery restore',
        10: 'Zone lost',
        11: 'Zone lost restore',
        12: 'Start of cross time',
    }

    NONE_EVENT_CODES = {
        17: 'Special expansion event',
        18: 'Duress',
        19: 'Manual fire',
        20: 'Auxiliary 2 panic',
        22: 'Panic',
        23: 'Keypad tamper',
        34: 'Telephone fault',
        35: 'Telephone fault restore',
        38: 'Fail to communicate',
        39: 'Log full',
        44: 'Auto-test',
        45: 'Start program',
        46: 'End program',
        47: 'Start download',
        48: 'End download',
        50: 'Ground fault',
        51: 'Ground fault restore',
        52: 'Manual test',
        54: 'Start of listen in',
        55: 'Technician on site',
        56: 'Technician left',
        57: 'Control power up',
        119: 'Time set',
        123: 'Begin walk-test',
        124: 'End walk-test',
        125: 'Re-exit',
        127: 'Data lost',
    }

    DEVICE_EVENT_CODES = {
        24: 'Control box tamper',
        25: 'Control box tamper restore',
        26: 'AC fail',
        27: 'AC fail restore',
        28: 'Low battery',
        29: 'Low battery restore',
        30: 'Over current',
        31: 'Over current restore',
        32: 'Siren tamper',
        33: 'Siren tamper restore',
        36: 'Expander trouble',
        37: 'Expander trouble restore',
    }

    USER_EVENT_CODES = {
        40: 'Opening',
        41: 'Closing',
        42: 'Exit error',
        43: 'Recent closing',
        49: 'Cancel',
        53: 'Closed with zones bypassed',
        120: 'First to open',
        121: 'Last to close',
        122: 'PIN entered with bit 7 set',
        126: 'Output trip',
    }

    def __init__(self):
        self.number = 0
        self.log_size = 0
        self.event_type = 119
        self.reportable = None
        self.zone_user_device = 0
        self.partition_number = 0
        self.timestamp = None

    @property
    def event(self):
        for codes in (self.ZONE_EVENT_CODES, self.USER_EVENT_CODES,
                      self.DEVICE_EVENT_CODES, self.NONE_EVENT_CODES):
            if self.event_type in codes:
                return codes[self.event_type]
        return 'Unknown event %i' % self.event_type

    @property
    def event_string(self):
        if self.event_type in self.ZONE_EVENT_CODES:
            return 'Zone %i %s' % (self.zone_user_device, self.event)
        if self.event_type in self.USER_EVENT_CODES:
            return 'User %i %s' % (self.zone_user_device, self.event)
        if self.event_type in self.DEVICE_EVENT_CODES:
            return 'Device %i %s' % (self.zone_user_device, self.event)
        if self.event_type in self.NONE_EVENT_CODES:
            return self.event
        return 'Unknown event %i for target %i' % (self.event_type,
                                                   self.zone_user_device)


class User(object):
    AUTHORITY_FLAGS = (
        ['Reserved', 'Arm only', 'Arm only (during close window)',
         'Master / Program', 'Arm / Disarm', 'Bypass enable',
         'Open / close report enable'],
        ['Output 1 enable', 'Output 2 enable', 'Output 3 enable',
         'Output 4 enable', 'Arm / disarm', 'Bypass enable',
         'Open / close report enable'],
    )

    def __init__(self, number):
        self.number = number
        self.pin = []
        self.authority_flags = []
        self.authorized_partitions = []


class NX584Extension(object):
    def __init__(self, controller):
        self._controller = controller

    def zone_status(self, zone):
        pass

    def partition_status(self, partition):
        pass

    def device_command(self, house, unit, command):
        pass

    def system_status(self, system):
        pass

    def log_event(self, event):
        pass



import json
import requests
import time


class Client(object):
    def __init__(self, url):
        self._url = url
        self._session = requests.Session()
        self._last_event_index = 0

    def list_zones(self):
        r = self._session.get(self._url + '/zones')
        try:
            return r.json['zones']
        except TypeError:
            return r.json()['zones']

    def list_partitions(self):
        r = self._session.get(self._url + '/partitions')
        try:
            return r.json['partitions']
        except TypeError:
            return r.json()['partitions']

    def arm(self, armtype='auto', partition=1):
        if armtype not in ['stay', 'exit', 'auto']:
            raise Exception('Invalid arm type')
        r = self._session.get(
            self._url + '/command',
            params={'cmd': 'arm',
                    'type': armtype,
                    'partition': partition})
        return r.status_code == 200

    def disarm(self, master_pin, partition=1):
        r = self._session.get(
            self._url + '/command',
            params={'cmd': 'disarm',
                    'master_pin': master_pin,
                    'partition': partition})
        return r.status_code == 200

    def set_bypass(self, zone, bypass):
        data = {'bypassed': bypass}
        r = self._session.put(self._url + '/zones/%i' % zone,
                              data=json.dumps(data),
                              headers={'Content-Type': 'application/json'})
        return r.status_code == 200

    def get_user(self, master_pin, user_number):
        params = {}
        while True:
            r = self._session.get(self._url + '/users/%i' % user_number,
                                  params=params,
                                  headers={'Master-Pin': master_pin})
            if r.status_code == 202:
                params['retry'] = 'yes'
                time.sleep(1)
                continue
            if r.status_code == 404:
                time.sleep(1)
                continue
            if r.status_code == 200:
                return r.json()
            print('Status code %i' % r.status_code)
            break

    def put_user(self, master_pin, user):
        cur_user = self.get_user(master_pin, user['number'])
        if not cur_user:
            return None
        r = self._session.put(self._url + '/users/%i' % user['number'],
                              headers={'Master-Pin': master_pin,
                                       'Content-Type': 'application/json'},
                              data=json.dumps(user))
        if r.status_code == 200:
            return r.json()

    def get_events(self, index=None, timeout=None):
        if index is None:
            index = self._last_event_index
        if timeout is None:
            timeout = 60
        r = self._session.get(self._url + '/events',
                              params={'index': index,
                                      'timeout': timeout})
        if r.status_code == 200:
            data = r.json()
            self._last_event_index = data['index']
            return data['events']

    def get_version(self):
        r = self._session.get(self._url + '/version')
        if r.status_code == 404:
            return '1.0'
        else:
            return r.json()['version']

    def get_info(self):
        r = self._session.get(self._url + '/version')
        if r.status_code == 404:
            return '1.0'
        else:
            return r.json()



import argparse
import logging
import logging.handlers
import os
import threading

from nx584 import api
from nx584 import controller

LOG_FORMAT = '%(asctime)-15s %(module)s %(levelname)s %(message)s'


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', default='config.ini',
                        metavar='FILE',
                        help='Path to config file')
    parser.add_argument('--debug', default=False, action='store_true',
                        help='Enable debug')
    parser.add_argument('--debug-log', default=None,
                        help='Debug to this log file only')
    parser.add_argument('--log', default=None,
                        metavar='FILE',
                        help='Path to log file')
    parser.add_argument('--connect', default=None,
                        metavar='HOST:PORT',
                        help='Host and port to connect for serial stream')
    parser.add_argument('--serial', default=None,
                        metavar='PORT',
                        help='Serial port to open for stream')
    parser.add_argument('--baudrate', default=38400, type=int,
                        metavar='BAUD',
                        help='Serial baudrate')
    parser.add_argument('--listen', default='127.0.0.1',
                        metavar='ADDR',
                        help='Listen address (defaults to 127.0.0.1)')
    parser.add_argument('--port', default=5007, type=int,
                        help='Listen port (defaults to 5007)')
    args = parser.parse_args()

    LOG = logging.getLogger()
    LOG.setLevel(logging.DEBUG)
    formatter = logging.Formatter(LOG_FORMAT)
    istty = os.isatty(0)

    if (args.debug or args.debug_log) and not istty:
        debug_handler = logging.handlers.RotatingFileHandler(
            args.debug_log or 'debug.log',
            maxBytes=1024*1024*10,
            backupCount=3)
        debug_handler.setFormatter(formatter)
        debug_handler.setLevel(logging.DEBUG)
        LOG.addHandler(debug_handler)

    if istty or not args.log:
        verbose_handler = logging.StreamHandler()
        verbose_handler.setFormatter(formatter)
        verbose_handler.setLevel(
            (args.debug and not args.debug_log) and logging.DEBUG or
            logging.INFO)
        LOG.addHandler(verbose_handler)

    if args.log:
        log_handler = logging.handlers.RotatingFileHandler(
            args.log,
            maxBytes=1024*1024*10,
            backupCount=3)
        log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.INFO)
        LOG.addHandler(log_handler)

    LOG.info('Ready')
    logging.getLogger('connectionpool').setLevel(logging.WARNING)

    if args.connect:
        host, port = args.connect.split(':')
        ctrl = controller.NXController((host, int(port)),
                                       args.config)
    elif args.serial:
        ctrl = controller.NXController((args.serial, args.baudrate),
                                       args.config)
    else:
        LOG.error('Either host:port or serial and baudrate are required')
        return

    api.CONTROLLER = ctrl

    t = threading.Thread(target=ctrl.controller_loop_safe)
    t.daemon = True
    t.start()

    api.app.run(debug=False, host=args.listen, port=args.port, threaded=True)



import flask
import json
import logging


LOG = logging.getLogger('api')
CONTROLLER = None
app = flask.Flask('nx584')


def show_zone(zone):
    return {
        'number': zone.number,
        'name': zone.name,
        'state': zone.state,
        'bypassed': zone.bypassed,
        'condition_flags': zone.condition_flags,
        'type_flags': zone.type_flags,
    }


def show_partition(partition):
    return {
        'number': partition.number,
        'condition_flags': partition.condition_flags,
        'armed': 'Armed' in partition.condition_flags,
        'last_user': partition.last_user,
    }


def show_user(user):
    if all([x > 9 for x in user.pin]):
        pin = None
    else:
        pin = ''.join([str(c) if c < 10 else '' for c in user.pin])
    return {
        'number': user.number,
        'pin': pin,
        'authority_flags': user.authority_flags,
        'authorized_partitions': user.authorized_partitions,
    }


@app.route('/zones')
def index_zones():
    try:
        result = json.dumps({
            'zones': [show_zone(zone) for zone in CONTROLLER.zones.values()]})
        return flask.Response(result,
                              mimetype='application/json')
    except Exception as e:
        LOG.exception('Failed to index zones')


@app.route('/partitions')
def index_partitions():
    try:
        result = json.dumps({
            'partitions': [show_partition(partition)
                           for partition in CONTROLLER.partitions.values()]})
        return flask.Response(result,
                              mimetype='application/json')
    except Exception as e:
        LOG.exception('Failed to index partitions')


@app.route('/command')
def command():
    args = flask.request.args
    if args.get('cmd') == 'arm':
        if args.get('type') == 'stay':
            CONTROLLER.arm_stay(int(args.get('partition', 1)))
        elif args.get('type') == 'exit':
            CONTROLLER.arm_exit(int(args.get('partition', 1)))
        else:
            CONTROLLER.arm_auto(int(args.get('partition', 1)))
    elif args.get('cmd') == 'disarm':
        CONTROLLER.disarm(args.get('master_pin'), int(args.get('partition', 1)))
    return flask.Response()


@app.route('/zones/<int:zone>', methods=['PUT'])
def put_zone(zone):
    zone = CONTROLLER.zones.get(zone)
    if not zone:
        flask.abort(404)
    zonedata = flask.request.json
    if 'bypassed' in zonedata:
        want_bypass = zonedata['bypassed']
        if want_bypass == zone.bypassed:
            flask.abort(409)
        CONTROLLER.zone_bypass_toggle(zone.number)
    result = json.dumps(show_zone(zone))
    return flask.Response(result,
                          mimetype='application/json')


@app.route('/users/<int:user>')
def get_user(user):
    args = flask.request.args
    master_pin = flask.request.headers.get('Master-Pin')
    if not master_pin:
        return 'Master PIN required', 403
    if user not in CONTROLLER.users:
        if 'retry' not in args:
            CONTROLLER.get_user_info(master_pin, user)
            return '', 202
        else:
            return 'Not Found', 404

    user = CONTROLLER.users[user]
    result = json.dumps(show_user(user))
    return flask.Response(result,
                          mimetype='application/json')


@app.route('/users/<int:user>', methods=['PUT'])
def put_user(user):
    if user == 1:
        return 'I refuse to let you break your master user', 403
    master_pin = flask.request.headers.get('Master-Pin')
    if not master_pin:
        return 'Master PIN required', 403
    if user not in CONTROLLER.users:
        CONTROLLER.get_user_info(master_pin, user)
        return '', 204

    user = CONTROLLER.users[user]
    if 'master' in ''.join(user.authority_flags).lower():
        return 'I refuse to let you break a master user', 403

    userdata = flask.request.json
    changed = []
    if 'pin' in userdata:
        pin = userdata['pin']
        changed.append('pin')
        if pin is None:
            user.pin = [15] * 6
        elif len(pin) == 4:
            user.pin = [int(i) for i in pin] + [15, 15]
        elif len(pin) == 6:
            user.pin = [int(i) for i in pin]
        else:
            return 'Invalid PIN format', 400

    if changed:
        CONTROLLER.set_user_info(master_pin, user, changed)

    return flask.Response(json.dumps(show_user(user)),
                          mimetype='application/json')


@app.route('/events')
def get_events():
    index = int(flask.request.args.get('index', 0))
    timeout = int(flask.request.args.get('timeout', 10))
    events = CONTROLLER.event_queue.get(index, timeout=timeout)
    if events:
        index = events[-1].number
        events = [event.payload for event in events]
    return flask.Response(json.dumps({'events': events,
                                      'index': index}),
                          mimetype='application/json')


@app.route('/version')
def get_version():
    return flask.Response(json.dumps(
        {'version': '1.2',
         'last_active': int(CONTROLLER.last_active)}),
                          mimetype='application/json')



import threading


class Event(object):
    def __init__(self, number, payload):
        self.number = number
        self.payload = payload

    def __repr__(self):
        return 'Event<%i>' % self.number


class EventQueue(object):
    def __init__(self, length, start=0):
        self._queue = []
        self._queue_lock = threading.Lock()
        self._length = length
        self._index = start
        self._condition = threading.Condition()
        self._min = start
        self._max = start

    def push(self, thing):
        self._condition.acquire()
        self._max += 1
        self._queue.append(Event(self._max, thing))
        self._queue = self._queue[0 - self._length:]
        self._min = self._queue[0].number
        self._condition.notify_all()
        self._condition.release()

    @property
    def current(self):
        return self._max

    def get(self, index, timeout=None):
        self._condition.acquire()

        data_available = lambda: index < self._min or self._max > index

        if not data_available():
            self._condition.wait(timeout)
        the_index = index - self._min + 1
        if the_index < 0:
            the_index = 0
        if data_available():
            result = self._queue[the_index:]
        else:
            result = None
        self._condition.release()
        return result



from setuptools import setup

setup(name='pynx584',
      version='0.8',
      description='NX584/NX8E Interface Library and Server',
      author='Dan Smith',
      author_email='dsmith+nx584@danplanet.com',
      url='http://github.com/kk7ds/pynx584',
      packages=['nx584'],
      install_requires=['requests', 'stevedore', 'prettytable', 'pyserial', 'flask'],
      scripts=['nx584_server', 'nx584_client'],
  )



try:
    import ConfigParser as configparser
except ImportError:
    import configparser
import datetime
import logging
import serial
import socket
import time

import stevedore.extension

from nx584 import event_queue
from nx584 import mail
from nx584 import model


LOG = logging.getLogger('controller')


def parse_ascii(data):
    data_bytes = []
    for i in range(0, len(data), 2):
        data_bytes.append(int(data[i:i + 2], 16))
    return data_bytes


def make_ascii(data):
    data_chars = []
    for b in data:
        data_chars.append('%02X' % b)
    return ''.join(data_chars)


def make_pin_buffer(digits):
    pinbuf = []
    for i in range(0, 6, 2):
        try:
            pinbuf.append((int(digits[i + 1]) << 4) | int(digits[i]))
        except (IndexError, TypeError):
            pinbuf.append(0xFF)
    return pinbuf


def fletcher(data, k=16):
    if  k not in (16, 32, 64):
        raise ValueError("Valid choices of k are 16, 32 and 64")
    nbytes = k // 16
    mod = 2 ** (8 * nbytes) - 1
    s = s2 = 0
    for t in data:
        s += t
        s2 += s
    cksum = s % mod + (mod + 1) * (s2 % mod)
    sum1 = cksum & 0xFF
    sum2 = (cksum & 0xFF00) >> 8
    return sum1, sum2


class NXFrame(object):
    def __init__(self):
        self.length = 0
        self.msgtype = 0
        self.checksum = 0
        self.data = []
        self.ackreq = False

    @property
    def ack_required(self):
        return self.ackreq

    @staticmethod
    def decode_line(line_bytes):
        self = NXFrame()

        s1, s2 = fletcher(line_bytes[:-2])
        if [s1, s2] != line_bytes[-2:]:
            raise ReadFailure('Checksum mismatch on received frame')

        self.length = line_bytes[0]
        msgtypefield = line_bytes[1]
        self.checksum = (line_bytes[-2] << 8) & line_bytes[-1]
        self.data = line_bytes[2:-2]
        self.ackreq = bool(msgtypefield & 0x80)
        self.msgtype = msgtypefield & 0x7F
        return self

    @property
    def type_name(self):
        return model.MSG_TYPES[self.msgtype]


class ReadTimeout(Exception):
    pass


class ConnectionLost(Exception):
    pass


class ReadFailure(Exception):
    pass


class NXProtocol(object):
    """Abstract the act of talking to the control panel."""
    def __init__(self, stream):
        self.stream = stream

    def read(self, n):
        """Read bytes from the stream.

        Does the right thing for sockets and serial ports.

        :raises: ReadTimeout if there is no data
        :raises: ConnectionLost if something goes wrong
        """
        if hasattr(self.stream, 'read'):
            r = self.stream.read(n)
            if r == b'':
                raise ReadTimeout()
            return r
        elif hasattr(self.stream, 'recv'):
            try:
                r = self.stream.recv(n)
            except socket.timeout:
                raise ReadTimeout()
            except (socket.error, OSError):
                raise ConnectionLost()
            if r == b'':
                raise ConnectionLost()
            return r
        else:
            raise ConnectionLost('What is this stream?')

    def write(self, data):
        """Write bytes to the stream.

        :raises: ConnectionLost if something goes wrong
        """
        if hasattr(self.stream, 'write'):
            self.stream.write(data)
        elif hasattr(self.stream, 'send'):
            try:
                self.stream.send(data)
            except (socket.error, OSError):
                raise ConnectionLost()
        else:
            raise ConnectionLost('What is this stream?')

    def discard_until(self, byte):
        """Reads any unexpected data until we encounter an expected byte.

        :raises: ReadTimeout (via self.read()) if we never get it.
        """
        while True:
            c = self.read(1)
            if c == byte:
                return c
            LOG.warning('Seeking (discarded %s %02x)' % (c, c[0]))

    def read_frame(self):
        """Read a whole frame from the stream.

        :returns: An array of int values, inclusive of the length and checksum
                  bytes
        :raises: ConnectionLost if something goes wrong or the stream cannot
                 be parsed
        """
        return []

    def write_frame(self, data):
        """Write a whole frame to the stream.

        NOTE: This adds the length and checksum bytes automatically.

        :param: data is an array of integers
        """
        pass


class NXASCII(NXProtocol):
    def read_frame(self):
        self.discard_until(b'\n')

        start = time.time()
        line = ''
        while not line.endswith('\r'):
            try:
                c = self.read(1).decode()
            except ReadTimeout:
                LOG.error('Mid-frame read timeout (got %i: %r)' % (
                    len(line), line))
                raise ConnectionLost()
            line += c
            if time.time() - start > 60:
                LOG.error('Timeout reading a line, killing connection')
                raise ConnectionLost()

        LOG.debug('Parsing ASCII frame %r' % line.strip())
        try:
            return parse_ascii(line.strip())
        except Exception:
            LOG.exception('Failed to parse raw ASCII line %r' % line.strip())
            raise ConnectionLost()

    def write_frame(self, data):
        data = [len(data)] + data
        data += fletcher(data)
        raw = '\n%s\r' % make_ascii(data)
        self.write(raw.encode())


class NXBinary(NXProtocol):
    def read_frame(self):
        self.discard_until(b'\x7e')

        length = self.read(1)[0]
        buffer = [length]
        start = time.time()

        bytestuffed = False

        i = 1
        while (i <= length + 2):
            try:
                c = self.read(1)[0]
            except ReadTimeout:
                LOG.error('Mid-frame read timeout (expected %i got %i)' % (
                    length, i))
                raise ConnectionLost()

            if c == 0x7e:
                raise ConnectionLost('Received start byte mid-frame!')

            # Adjust any byte stuffed bytes - Skip the current byte if so,
            # then XOR the following byte with 0x20
            if c == 0x7d:
                bytestuffed = True
            else:
                if bytestuffed:
                    c = c ^ 0x20
                    bytestuffed = False
                buffer.append(c)
                i += 1

            if time.time() - start > 60:
                LOG.error('Timeout reading a line, killing connection')
                raise ConnectionLost()

        return buffer

    def write_frame(self, data):
        data = [len(data)] + data
        data += fletcher(data)
        # Byte Stuff Any 0x7d, 0x7e bytes
        bytestuff = [i for i, x in enumerate(data) if x == 0x7d]
        for i, index in reversed(list(enumerate(bytestuff))):
            data[index:index+1] = [0x7d, 0x5d]
        bytestuff = [i for i, x in enumerate(data) if x == 0x7e]
        for i, index in reversed(list(enumerate(bytestuff))):
            data[index:index+1] = [0x7d, 0x5e]

        # add the start character 0x7e
        data.insert(0, 0x7e)

        self.write(bytes(data))


class StreamWrapper(object):
    """Wraps a connection to the control panel.

    This mostly just manages connecting, reconnecting, and some error handling,
    transparent to the protocol (ASCII or Binary) being used.
    """
    def __init__(self, portspec, config):
        self._portspec = portspec
        self._config = config

        try:
            self._use_binary_protocol = self._config.getboolean(
                'config', 'use_binary_protocol')
        except configparser.NoOptionError:
            self._use_binary_protocol = False

        self._s = None
        self.protocol = None
        self.connect()

    def connect(self):
        sleep_time = 0
        while True:
            connected = self._connect()
            if connected:
                if self._use_binary_protocol:
                    self.protocol = NXBinary(self._s)
                else:
                    self.protocol = NXASCII(self._s)
                LOG.info('Connected')
                return True
            sleep_time = min(60, sleep_time + 5)
            LOG.debug('Waiting %i sec before retry...' % sleep_time)
            time.sleep(sleep_time)

    def reconnect(self):
        self._s.close()
        time.sleep(10)
        self.connect()

    def read_frame_raw(self):
        """Read a raw frame from the stream.

        Returns the raw bytes (inclusive of length and checksum) or None
        if there is no data to read.
        """
        try:
            return self.protocol.read_frame()
        except ConnectionLost as e:
            LOG.warning('Connection terminated: %s' % e)
            self.reconnect()
            return None
        except ReadTimeout:
            return None
        except UnicodeDecodeError as e:
            LOG.error('Failed to decode a line; reconnecting: %s' % e)
            return None

    def write_frame_raw(self, data):
        """Writes a frame to the stream.

        NOTE: length and checksum values are added automatically.
        """
        try:
            self.protocol.write_frame(data)
        except ConnectionLost:
            LOG.warning('Failed to send frame; reconnecting')
            self.reconnect()
            # Try to re-send if we reconnect so we don't lose this event
            self.protocol.write(data)


class SocketWrapper(StreamWrapper):
    def _connect(self):
        if self._s:
            self._s.close()
        try:
            self._s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            LOG.debug('Connecting...')
            self._s.connect(self._portspec)
            self._s.settimeout(0.5)
            return True
        except (socket.error, OSError) as ex:
            LOG.error('Failed to connect: %s' % ex)
            self._s = None
            self.protocol = None
            return False


class SerialWrapper(StreamWrapper):
    def _connect(self):
        port, baudrate = self._portspec
        self._s = serial.Serial(port, baudrate, timeout=0.25)
        return self._s.isOpen();


class NXController(object):
    def __init__(self, portspec, configfile):
        self._portspec = portspec
        self._configfile = configfile
        self._queue_waiting = False
        self._queue_should_wait = False
        self._queue = []
        self.last_active = 0
        self.zones = {}
        self.partitions = {}
        self.users = {}
        self.system = model.System()
        ext_mgr = stevedore.extension.ExtensionManager(
            'pynx584', invoke_on_load=True, invoke_args=(self,))
        self.extensions = [ext_mgr[name] for name in ext_mgr.names()]
        LOG.info('Loaded extensions %s' % ext_mgr.names())
        self._load_config()
        self.event_queue = event_queue.EventQueue(100)
        try:
            self.zone_name_update = self._config.getboolean('config', 'zone_name_update')
        except configparser.NoOptionError:
            self.zone_name_update = True
        try:
            self._idle_time_heartbeat_seconds = self._config.getint(
                'config', 'idle_time_heartbeat_seconds')
        except configparser.NoOptionError:
            self._idle_time_heartbeat_seconds = 120

    def connect(self):
        if '/' in self._portspec[0] or 'COM' in self._portspec[0]:
            self._ser = SerialWrapper(self._portspec, self._config)
        else:
            self._ser = SocketWrapper(self._portspec, self._config)
            LOG.info('Connected')

    def _load_config(self):
        self._config = configparser.ConfigParser()
        self._config.read(self._configfile)

        if not self._config.has_section('config'):
            self._config.add_section('config')

        if self._config.has_section('zones'):
            for opt in self._config.options('zones'):
                number = int(opt)
                name = self._config.get('zones', opt)
                zone = self._get_zone(number)
                zone.name = name

    def _write_config(self):
        if not self._config.has_section('zones'):
            self._config.add_section('zones')

        for zone in self.zones.values():
            if (not self._config.has_option('zones', str(zone.number)) and
                    zone.name != 'Unknown'):
                self._config.set('zones', str(zone.number), zone.name)
        try:
            with open(self._configfile, 'w') as configfile:
                self._config.write(configfile)
        except IOError as ex:
            LOG.error('Unable to write %s: %s' % (self._configfile, ex))

    @property
    def interior_zones(self):
        return [x for x in self.zones.values() if 'Interior' in x.type_flags]

    @property
    def interior_bypassed(self):
        return all(['Inhibit' in x.condition_flags
                    for x in self.interior_zones])

    def process_next(self):
        data = self._ser.read_frame_raw()
        if not data:
            return None
        self.last_active = time.time()
        try:
            frame = NXFrame.decode_line(data)
        except ReadFailure as e:
            LOG.error(str(e))
            return None
        return frame

    def _send(self, data):
        try:
            self._ser.write_frame_raw(data)
        except Exception:
            LOG.exception('Failed to send frame %r' % data)

    def send_ack(self):
        self._send([0x1D])

    def send_nack(self):
        self._send([0x1E])

    def get_zone_name(self, number):
        self._queue.append([0x23, number - 1])

    def get_zone_status(self, number):
        self._queue.append([0x24, number - 1])

    def arm_stay(self, partition):
        self._queue.append([0x3E, 0x00, partition])

    def arm_exit(self, partition):
        self._queue.append([0x3E, 0x02, partition])

    def arm_auto(self, partition):
        self._queue.append([0x3D, 0x05, 0x01, 0x01])

    def disarm(self, master_pin, partition):
        self._queue.append([0x3C] +
                           make_pin_buffer(master_pin) +
                           [0x01, partition])

    def zone_bypass_toggle(self, zone):
        self._queue.append([0x3F, zone - 1])

    def get_system_status(self):
        self._queue.append([0x28])

    def get_partition_status(self, partition):
        self._queue.append([0x26, partition - 1])

    def set_time(self):
        now = datetime.datetime.now()
        self._queue.append([0x3B,
                            now.year - 2000,
                            now.month,
                            now.day,
                            now.hour,
                            now.minute,
                            ((now.weekday() + 1) % 7) + 1])

    def get_user_info(self, master_pin, user_number):
        if len(master_pin) < 6:
            master_pin += '00'
        if len(master_pin) != 6:
            LOG.error('Master pin %r incorrect length' % master_pin)
            return False
        digits = make_pin_buffer(master_pin)
        self._queue.append([0x32] + digits + [user_number])
        LOG.debug('Sending for user info %s' % digits)
        return True

    def set_user_info(self, master_pin, user, changed):
        if user.number < 1:
            LOG.error('Unable to set PIN for user %i' % user.number)
            return False
        if 'pin' in changed:
            mstr_digits = make_pin_buffer(master_pin)
            user_digits = make_pin_buffer(user.pin)
            LOG.info('Setting user %i PIN to `%s`' % (
                user.number,
                ''.join(str(x) for x in user.pin if x < 10)))
            self._queue.append(
                [0x34] + mstr_digits + [user.number] + user_digits)
        return True

    def _get_zone(self, number):
        if number not in self.zones:
            self.zones[number] = model.Zone(number)
        return self.zones[number]

    def _get_partition(self, number):
        if number not in self.partitions:
            self.partitions[number] = model.Partition(number)
        return self.partitions[number]

    def _get_user(self, number):
        if number not in self.users:
            self.users[number] = model.User(number)
        return self.users[number]

    def process_msg_3(self, frame):
        # Zone Name
        number = frame.data[0] + 1
        name = ''.join([chr(x) for x in frame.data[1:]])
        LOG.info('Zone %i: %s' % (number, repr(name.strip())))
        if self.zone_name_update:
            self._get_zone(number).name = name.strip()
            LOG.debug('Zone info from %s' % self.zones.keys())
            self._write_config()
        else:
            LOG.debug('Zone name not updating')

    def process_msg_4(self, frame):
        # Zone Status
        zone = self._get_zone(frame.data[0] + 1)
        condition = frame.data[5]
        types = frame.data[2:5]
        zone.state = bool(condition & 0x01)

        zone.condition_flags = []
        for index, string in enumerate(model.Zone.STATUS_FLAGS):
            if condition & (1 << index):
                zone.condition_flags.append(string)

        zone.type_flags = []
        for byte, flags in enumerate(model.Zone.TYPE_FLAGS):
            for bit, name in enumerate(flags):
                if types[byte] & (1 << bit):
                    zone.type_flags.append(name)

        LOG.info('Zone %i (%s) state is %s' % (
            zone.number, zone.name,
            zone.state and 'FAULT' or 'NORMAL'))
        LOG.debug('Zone %i (%s) %s %s' % (zone.number, zone.name,
                                          zone.condition_flags,
                                          zone.type_flags))
        event = {'type': 'zone_status',
                 'timestamp': datetime.datetime.now().isoformat(),
                 'zone': zone.number,
                 'zone_state': zone.state,
                 'zone_flags': zone.condition_flags,
             }
        self.event_queue.push(event)
        for ext in self.extensions:
            ext.obj.zone_status(zone)

    def _send_flag_notifications(self, flags_section, flags_key, asserted,
                                 deasserted, email_fn):
        changed = asserted | deasserted
        try:
            flags = set(self._config.get(flags_section,
                                         flags_key).split(','))
        except (configparser.NoOptionError, configparser.NoSectionError):
            flags = set([])

        if any([flag in changed for flag in flags]):
            if deasserted & flags:
                status = '(restored)'
            else:
                status = ''
            sub = '%s %s' % (list(changed & flags), status)
            msg = ('Asserted: %s\n' % (', '.join(asserted)) +
                   'Deasserted: %s\n' % (', '.join(deasserted)))
            email_fn(sub, msg)

    def process_msg_6(self, frame):
        partition = self._get_partition(frame.data[0] + 1)
        partition.last_user = frame.data[5]
        types = frame.data[1:5] + frame.data[6:8]
        was_armed = partition.armed
        orig_flags = partition.condition_flags
        partition.condition_flags = []
        for byte, flags in enumerate(model.Partition.CONDITION_FLAGS):
            for bit, name in enumerate(flags):
                if types[byte] & (1 << bit):
                    partition.condition_flags.append(name)
        if was_armed != partition.armed:
            LOG.info('Partition %i %s armed' % (
                partition.number,
                '' if partition.armed else 'not'))
        LOG.debug('Partition %i %s' % (partition.number,
                                       partition.condition_flags))
        for ext in self.extensions:
            ext.obj.partition_status(partition)

        deasserted = set(orig_flags) - set(partition.condition_flags)
        asserted = set(partition.condition_flags) - set(orig_flags)
        changed = asserted | deasserted

        event = {'type': 'partition',
                 'timestamp': datetime.datetime.now().isoformat(),
                 'partition': partition.number,
                 'partition_flags_asserted': list(asserted),
                 'partition_flags': partition.condition_flags,
                 'partition_was_armed': was_armed,
                 'partition_is_armed': partition.armed,
             }
        self.event_queue.push(event)

        if changed:
            mail.send_partition_email(self._config, partition,
                                      deasserted, asserted)

            def email_status(sub, msg):
                mail.send_partition_status_email(self._config, partition,
                                                 'status', sub, msg)

            def email_alarms(sub, msg):
                mail.send_partition_status_email(self._config, partition,
                                                 'alarms', sub, msg)

            section = 'partition_%i' % partition.number
            self._send_flag_notifications(section, 'status_flags', asserted,
                                          deasserted, email_status)
            self._send_flag_notifications(section, 'alarm_flags', asserted,
                                          deasserted, email_alarms)


    def process_msg_8(self, frame):
        errors = model.System.STATUS_FLAGS[1] + model.System.STATUS_FLAGS[2]
        status = frame.data[1:10]
        self.system.panel_id = frame.data[0]
        orig_flags = self.system.status_flags
        self.system.status_flags = []
        for byte, flags in enumerate(model.System.STATUS_FLAGS):
            for bit, name in enumerate(flags):
                if status[byte] & (1 << bit):
                    self.system.status_flags.append(name)
        LOG.debug('System status received (panel id 0x%02x)' % (
            self.system.panel_id))

        def _log(flag, asserted):
            if flag not in errors:
                fn = LOG.info
            else:
                if asserted:
                    fn = LOG.error
                else:
                    fn = LOG.warn
            if asserted:
                pfx = ''
            else:
                pfx = 'de-'
            msg = 'System %sasserts %s' % (pfx, flag)
            fn(msg)

        deasserted = set(orig_flags) - set(self.system.status_flags)
        asserted = set(self.system.status_flags) - set(orig_flags)

        for flag in deasserted:
            _log(flag, False)
        for flag in asserted:
            _log(flag, True)

        for ext in self.extensions:
            ext.obj.system_status(self.system)

        if asserted or deasserted:
            mail.send_system_email(self._config, deasserted, asserted)

        for i in range(1, 9):
            if ('Valid partition %i' % i) in asserted:
                LOG.debug('Requesting partition status for %i' % i)
                self.get_partition_status(i)

    def process_msg_9(self, frame):
        commands = {0x28: 'on', 0x38: 'off'}
        house = chr(ord('A') + frame.data[0])
        unit = frame.data[1]
        cmd = commands.get(frame.data[2], frame.data[2])
        LOG.info('Device %s%02i command %s' % (house, unit, cmd))
        event = {'type': 'device-command',
                 'timestamp': datetime.datetime.now().isoformat(),
                 'device': '%s%02i' % (house, unit),
                 'command': '%s' % cmd}
        self.event_queue.push(event)
        for ext in self.extensions:
            ext.obj.device_command(house, unit, cmd)

    def process_msg_10(self, frame):
        event = model.LogEvent()
        event.number = frame.data[0]
        event.log_size = frame.data[1]
        event.event_type = frame.data[2] & 0x7F
        event.reportable = bool(frame.data[2] & 0x80)
        event.zone_user_device = frame.data[3]
        event.partition_number = frame.data[4]
        euro_format = self._config.getboolean('config', 'euro_date_format',
                                              fallback=False)
        if euro_format:
            month = frame.data[6]
            day = frame.data[5]
        else:
            month = frame.data[5]
            day = frame.data[6]

        hour = frame.data[7]
        minute = frame.data[8]
        now = datetime.datetime.now()
        if month > now.month:
            year = now.year - 1
        else:
            year = now.year
        try:
            event.timestamp = datetime.datetime(
                year=year, month=month, day=day,
                hour=hour, minute=minute)
        except ValueError:
            LOG.error('Log event had invalid date, or format needs to be set')
            return
        LOG.info('Log event: %s at %s' % (event.event_string,
                                          event.timestamp))
        _event = {'type': 'log',
                  'event': event.event_string,
                  'timestamp': event.timestamp.isoformat(),
              }
        self.event_queue.push(_event)
        for ext in self.extensions:
            ext.obj.log_event(event)

        mail.send_log_event_mail(self._config, event)

    def process_msg_18(self, frame):
        user = self._get_user(frame.data[0])
        user.pin = []
        user.authority_flags = []
        user.authorized_partitions = []
        for byte in frame.data[1:4]:
            user.pin.append(byte & 0x0F)
            user.pin.append((byte & 0xF0) >> 4)
        authbytetype = frame.data[4] & 0x80
        authbyte = frame.data[4] & 0x7F
        flags = model.User.AUTHORITY_FLAGS[1 if authbytetype else 0]
        for i, flag in enumerate(flags):
            if authbyte & (1 << i):
                user.authority_flags.append(flag)
        for i in range(0, 8):
            if frame.data[5] & (1 << i):
                user.authorized_partitions.append(i + 1)
        LOG.info('Received information about user %i' % user.number)

    def _run_queue(self):
        if not self._queue:
            return
        msg = self._queue.pop(0)
        LOG.debug('Sending queued %s' % msg)
        self._send(msg)

    def generate_heartbeat_activity(self):
        self.get_system_status()

    def controller_loop(self):
        self.set_time()
        self.get_system_status()

        try:
            max_zone = self._config.getint('config', 'max_zone')
        except configparser.NoOptionError:
            max_zone = 8
            self._config.set('config', 'max_zone', str(max_zone))

        for i in range(1, max_zone + 1):
            self.get_zone_status(i)
            if not self._config.has_option('zones', str(i)):
                self.get_zone_name(i)

        watchdog = time.time()

        while self.running:
            frame = self.process_next()
            if frame is None:
                if time.time() - watchdog < self._idle_time_heartbeat_seconds:
                    self._run_queue()
                else:
                    # After time with no activity - generate
                    # something to make sure we are still alive
                    LOG.info('No activity for a while, heartbeating')
                    self.generate_heartbeat_activity()
                    watchdog = time.time()
                continue
            watchdog = time.time()
            LOG.debug('Received: %i %s (data %s)' % (frame.msgtype,
                                                     frame.type_name,
                                                     frame.data))
            if frame.ack_required:
                LOG.debug('Sending ACK')
                self.send_ack()
            else:
                pass
                # This is sometimes too fast if we get two responses
                # from a single command (like time set). Need to keep
                # track of waiting for replies and re-send things when
                # we don't hear back.
                # self._run_queue()
            name = 'process_msg_%i' % frame.msgtype
            if hasattr(self, name):
                try:
                    getattr(self, name)(frame)
                except Exception as e:
                    LOG.exception('Failed to process message type %i',
                                  frame.msgtype)
            else:
                LOG.debug('Unsupported frame type %i (0x%02x)' % (
                    frame.msgtype, frame.msgtype))

    def controller_loop_safe(self):
        self.running = True

        while self.running:
            try:
                self.connect()
                self.controller_loop()
            except Exception as e:
                LOG.exception('Controller loop exited: %s' % e)
                LOG.warning('Waiting 10s before reconnecting...')
                time.sleep(10)
